"""
author: notjik
license: MIT License

copyright: (C) 2023 notjik
"""
from typing import Callable, Sequence

__all__ = ['Sort',
           'Find']


class Sort:
    @staticmethod
    # TODO: Bubble sorting method (https://sortvisualizer.com/bubblesort/)
    # TODO: Метод сортировки пузырьком (https://sortvisualizer.com/bubblesort/)
    def bubble(array: list[int | float | str | tuple | list] | tuple[int | float | str | tuple | list] | str, *,
               reverse: bool = False,
               alg: Callable[[int | float | str | tuple | list, int | float | str | tuple | list], bool] =
               lambda x, y: x < y) \
            -> list[int | float | str | tuple | list] | tuple[int | float | str | tuple | list] | str:
        tp = type(array)
        if not isinstance(array, list):
            array = list(array)
        for i in range(len(array) - 1):
            for j in range(len(array) - i - 1):
                if alg(array[j + 1], array[j]) if not reverse else alg(array[j], array[j + 1]):
                    array[j + 1], array[j] = array[j], array[j + 1]
        return ''.join(array) if tp == str else tp(array) if tp != set else list(array)

    @staticmethod
    # TODO: Shaker sorting method (https://sortvisualizer.com/shakersort/)
    # TODO: Метод сортировки шейкером (https://sortvisualizer.com/shakersort/)
    def shaker(array: list[int | float | str | tuple | list] | tuple[int | float | str | tuple | list] | str,
               reverse: bool = False,
               alg: Callable[[int | float | str | tuple | list, int | float | str | tuple | list], bool] =
               lambda x, y: x < y) \
            -> list[int | float | str | tuple | list] | tuple[int | float | str | tuple | list] | str:
        tp = type(array)
        if not isinstance(array, list):
            array = list(array)
        left, right = 0, len(array) - 1
        while left <= right:
            for i in range(right, left, -1):
                if alg(array[i], array[i - 1]) if not reverse else alg(array[i - 1], array[i]):
                    array[i - 1], array[i] = array[i], array[i - 1]
            left += 1
            for i in range(left, right):
                if alg(array[i + 1], array[i]) if not reverse else alg(array[i], array[i + 1]):
                    array[i], array[i + 1] = array[i + 1], array[i]
            right -= 1
        return ''.join(array) if tp == str else tp(array)

    @staticmethod
    # TODO: Insertion sorting method (https://sortvisualizer.com/insertionsort/)
    # TODO: Метод сортировки вставками (https://sortvisualizer.com/insertionsort/)
    def insertion(array: list[int | float | str | tuple | list] | tuple[int | float | str | tuple | list] | str, *,
                  reverse: bool = False,
                  alg: Callable[[int | float | str | tuple | list, int | float | str | tuple | list], bool] =
                  lambda x, y: x < y) \
            -> list[int | float | str | tuple | list] | tuple[int | float | str | tuple | list] | str:
        tp = type(array)
        if not isinstance(array, list):
            array = list(array)
        for i in range(1, len(array)):
            x = array[i]
            j = i - 1
            while j >= 0 and alg(x, array[j]) if not reverse else j >= 0 and alg(array[j], x):
                array[j + 1] = array[j]
                j -= 1
            array[j + 1] = x
        return ''.join(array) if tp == str else tp(array)

    @staticmethod
    # TODO: Selection sorting method (https://sortvisualizer.com/selectionsort/)
    # TODO: Метод сортировки выборкой (https://sortvisualizer.com/selectionsort/)
    def selection(array: list[int | float | str | tuple | list] | tuple[int | float | str | tuple | list] | str, *,
                  reverse: bool = False,
                  alg: Callable[[int | float | str | tuple | list, int | float | str | tuple | list], bool] =
                  lambda x, y: x < y) \
            -> list[int | float | str | tuple | list] | tuple[int | float | str | tuple | list] | str:
        tp = type(array)
        if not isinstance(array, list):
            array = list(array)
        for i in range(len(array)):
            tmp = i
            for j in range(i + 1, len(array)):
                if alg(array[j], array[tmp]) if not reverse else alg(array[tmp], array[j]):
                    tmp = j
            array[i], array[tmp] = array[tmp], array[i]
        return ''.join(array) if tp == str else tp(array)

    @staticmethod
    # TODO: Counting sorting method (https://www.cs.usfca.edu/~galles/visualization/CountingSort.html)
    # TODO: Метод сортировки подсчётом (https://www.cs.usfca.edu/~galles/visualization/CountingSort.html)
    def counting(array: list[int | float] | tuple[int | float], *,
                 reverse: bool = False) -> list[int | float] | tuple[int | float]:
        tp = type(array)
        array_init = [0] * (max(array) + 1)
        result = []
        for i in range(len(array)):
            array_init[array[i]] += 1
        for i in range(len(array_init)) if not reverse else range(len(array_init) - 1, -1, -1):
            if array_init[i]:
                for k in range(array_init[i]):
                    result.append(i)
        return tp(result)

    @staticmethod
    # TODO: Merge sorting method (https://sortvisualizer.com/mergesort/)
    # TODO: Метод сортировки слиянием (https://sortvisualizer.com/mergesort/)
    def merge(array: list[int | float | str | tuple | list] | tuple[int | float | str | tuple | list] | str, *,
              reverse: bool = False,
              alg: Callable[[int | float | str | tuple | list, int | float | str | tuple | list], bool] =
              lambda x, y: x < y) \
            -> list[int | float | str | tuple | list] | tuple[int | float | str | tuple | list] | str:
        tp = type(array)
        if not isinstance(array, list):
            array = list(array)
        if len(array) > 1:
            median = len(array) // 2
            left, right = array[:median], array[median:]
            Sort.merge(array=left, reverse=reverse, alg=alg)
            Sort.merge(array=right, reverse=reverse, alg=alg)
            i = j = k = 0
            while i < len(left) and j < len(right):
                if alg(left[i], right[j]) if not reverse else alg(right[j], left[i]):
                    array[k] = left[i]
                    i += 1
                else:
                    array[k] = right[j]
                    j += 1
                k += 1
            while i < len(left):
                array[k] = left[i]
                i += 1
                k += 1
            while j < len(right):
                array[k] = right[j]
                j += 1
                k += 1
        return ''.join(array) if tp == str else tp(array)

    @staticmethod
    # TODO: Hoare's sorting method (https://sortvisualizer.com/quicksort/)
    # TODO: Метод сортировки Хоара (https://sortvisualizer.com/quicksort/)
    def quick(array: list[int | float | str | tuple | list] | tuple[int | float | str | tuple | list] | str,
              start: int = 0,
              end: int | None = None, *,
              reverse: bool = False,
              alg: Callable[[int | float | str | tuple | list, int | float | str | tuple | list], bool] =
              lambda x, y: x < y) \
            -> list[int | float | str] | tuple[int | float | str] | str:
        def partition(array: list,
                      start: int,
                      end: int):
            pivot = start
            for i in range(start + 1, end + 1):
                if alg(array[i], array[start]) if not reverse else alg(array[start], array[i]):
                    pivot += 1
                    array[i], array[pivot] = array[pivot], array[i]
            array[pivot], array[start] = array[start], array[pivot]
            return pivot

        tp = type(array)
        if not isinstance(array, list):
            if isinstance(array, dict):
                raise ValueError('Dictionary is an invalid type for sorting')
            array = list(array)
        if end is None:
            end = len(array) - 1
        if start >= end:
            return []
        pivot = partition(array, start, end)
        Sort.quick(array=array, start=start, end=pivot - 1, alg=alg, reverse=reverse)
        Sort.quick(array=array, start=pivot + 1, end=end, alg=alg, reverse=reverse)
        return ''.join(array) if tp == str else tp(array)


class Find:
    @staticmethod
    # TODO: Linear search
    # TODO: Линейный поиск
    def linear(array: list[int | float | str | list | tuple] | tuple[int | float | str | list | tuple] | str,
               element: int | float | str | list | tuple, *,
               reverse: bool = False) -> int:
        for i in range(len(array)) if not reverse else range(len(array) - 1, -1, -1):
            if array[i] == element:
                return i
        return -1

    @staticmethod
    # TODO: Binary search (sorted collection)
    # TODO: Бинарный поиск (отсортированной коллекции)
    def binary(array: list[int | float | str | list | tuple] | tuple[int | float | str | list | tuple] | str,
               element: int | float | str | list | tuple,
               start: int = 0,
               end: int | None = None) -> int | None:
        if end is None:
            end = len(array) - 1
        i = -1
        while (start <= end) and (i == -1):
            median = (start + end) // 2
            if array[median] == element:
                i = median
            else:
                if element < array[median]:
                    end = median - 1
                else:
                    start = median + 1
        return i


if __name__ == '__main__':
    from timeit import timeit
    from random import randint

    l: int = 10
    array: list[int] = [randint(0, 1000) for r in range(l)]
    operations: int = 100

    print('Bubble:')
    print('Before: {}'.format(array))
    print('After: {}'.format(Sort.bubble(array[:])))
    print('{} operations per {}\n'.format(operations, timeit('Sort.bubble({})'.format(array[:]) if type(array) != str
                                                             else 'Sort.bubble("{}")'.format(array[:]),
                                                             'from __main__ import Sort', number=operations)))

    print('Shaker:')
    print('Before: {}'.format(array))
    print('After: {}'.format(Sort.shaker(array[:])))
    print('{} operations per {}\n'.format(operations, timeit('Sort.shaker({})'.format(array[:]) if type(array) != str
                                                             else 'Sort.shaker("{}")'.format(array[:]),
                                                             'from __main__ import Sort', number=operations)))

    print('Insertion:')
    print('Before: {}'.format(array))
    print('After: {}'.format(Sort.insertion(array[:])))
    print('{} operations per {}\n'.format(operations, timeit('Sort.insertion({})'.format(array[:]) if type(array) != str
                                                             else 'Sort.insertion("{}")'.format(array[:]),
                                                             'from __main__ import Sort', number=operations)))

    print('Selection:')
    print('Before: {}'.format(array))
    print('After: {}'.format(Sort.selection(array[:])))
    print('{} operations per {}\n'.format(operations, timeit('Sort.selection({})'.format(array[:]) if type(array) != str
                                                             else 'Sort.selection("{}")'.format(array[:]),
                                                             'from __main__ import Sort', number=operations)))

    if type(array) != str:
        print('Counting:')
        print('Before: {}'.format(array))
        print('After: {}'.format(Sort.counting(array[:])))
        print('{} operations per {}\n'.format(operations, timeit('Sort.counting({})'.format(array[:]),
                                                                 'from __main__ import Sort', number=operations)))

    print('Merge:')
    print('Before: {}'.format(array))
    print('After: {}'.format(Sort.merge(array[:])))
    print('{} operations per {}\n'.format(operations, timeit('Sort.merge({})'.format(array[:]) if type(array) != str
                                                             else 'Sort.merge("{}")'.format(array[:]),
                                                             'from __main__ import Sort', number=operations)))

    print('Quick:')
    print('Before: {}'.format(array))
    print('After: {}'.format(Sort.quick(array[:])))
    print('{} operations per {}'.format(operations, timeit('Sort.quick({})'.format(array[:]) if type(array) != str
                                                           else 'Sort.quick("{}")'.format(array[:]),
                                                           'from __main__ import Sort', number=operations)))



"""
author: notjik
license: MIT License

copyright: (C) 2023 notjik
"""
from time import perf_counter
from typing import Callable, Any

__all__ = ['NumbersProperties',
           'NumeralSystem',
           'PerformanceTracking']


class NumbersProperties:
    @staticmethod
    # TODO: Search for divisors
    # TODO: Поиск делителей
    def divisors(number: int, *, nontrivial: bool = False, prime: bool = False, integer: bool = False) -> list:
        if not isinstance(number, int) or isinstance(number, bool):
            raise TypeError('The current {} type is {}, expected int'.format(number, type(number).__name__))
        elif not isinstance(nontrivial, bool):
            raise TypeError('The current {} type is {}, expected bool'.format(nontrivial, type(nontrivial).__name__))
        elif not isinstance(prime, bool):
            raise TypeError('The current {} type is {}, expected bool'.format(prime, type(prime).__name__))
        elif not isinstance(integer, bool):
            raise TypeError('The current {} type is {}, expected bool'.format(integer, type(integer).__name__))
        elif number == 0:
            raise ZeroDivisionError('It is impossible to find divisors for zero')
        elif all((prime, nontrivial)):
            raise ValueError('It is impossible to find nontrivial prime divisors')
        elif all((prime, integer)):
            raise ValueError('It is impossible to find prime divisors in integers')
        else:
            result = []
        i = 1 if not (nontrivial or prime) else 2
        tmp_number = abs(number)
        if prime:
            while i ** 2 <= tmp_number:
                if tmp_number % i == 0:
                    result.append(i)
                    tmp_number //= i
                else:
                    i += 1
            if tmp_number > 1:
                result.append(tmp_number)
        else:
            while i ** 2 <= tmp_number:
                if i ** 2 == tmp_number:
                    result.append(i)
                elif tmp_number % i == 0:
                    result += [i, tmp_number // i]
                i += 1
        return result if not integer else result + list(map(lambda x: -x, result))

    @staticmethod
    # TODO: Finding the number of divisors (through the basic theorem of arithmetic)
    # TODO: Нахождение количества делителей (через основную теорему арифметики)
    def count_divisors(number: int, *, nontrivial: bool = False, prime: bool = False, integer: bool = False):
        if not isinstance(number, int) or isinstance(number, bool):
            raise TypeError('The current {} type is {}, expected int'.format(number, type(number).__name__))
        elif not isinstance(nontrivial, bool):
            raise TypeError('The current {} type is {}, expected bool'.format(nontrivial, type(nontrivial).__name__))
        elif not isinstance(prime, bool):
            raise TypeError('The current {} type is {}, expected bool'.format(prime, type(prime).__name__))
        elif not isinstance(integer, bool):
            raise TypeError('The current {} type is {}, expected bool'.format(integer, type(integer).__name__))
        elif number == 0:
            raise ZeroDivisionError('It is impossible to find divisors for zero')
        elif all((prime, nontrivial)):
            raise ValueError('It is impossible to find the count of nontrivial prime divisors')
        elif all((prime, integer)):
            raise ValueError('It is impossible to find the count prime divisors in integers')
        else:
            result = 1
        prime_divisors = NumbersProperties.divisors(abs(number), prime=True)
        if not prime:
            for i in set(prime_divisors):
                result *= prime_divisors.count(i) + 1
            if integer:
                result *= 2
            return result if not nontrivial else result - 2 if not integer else result - 4
        else:
            result = len(prime_divisors)
            return result

    @staticmethod
    # TODO: Finding the n-th term of the Fibonacci number (via the Binet's formula)
    # TODO: Нахождение n-го члена числа Фибоначчи (через формулу Бине)
    def fibonacci(number: int) -> int:
        if not isinstance(number, int) or isinstance(number, bool):
            raise TypeError('The current {} type is {}, expected int'.format(number, type(number).__name__))
        elif number < 0:
            raise ValueError('The n-th term of the sequence must be greater than 0')
        return round((((1 + 5 ** 0.5) / 2) ** number - ((1 - 5 ** 0.5) / 2) ** number) / 5 ** 0.5)

    @staticmethod
    # TODO: Checking a number for prime
    # TODO: Проверка числа на простоту
    def is_prime(number: int) -> bool:
        if not isinstance(number, int) or isinstance(number, bool):
            raise TypeError('The current {} type is {}, expected int'.format(number, type(number).__name__))
        if number < 2:
            return False
        i = 2
        while i * i <= number:
            if number % i == 0:
                return False
            i += 1
        return True

    @staticmethod
    # TODO: Checking the number by square
    # TODO: Проверка числа на квадрат
    def is_square(number: int) -> bool:
        if not isinstance(number, int) or isinstance(number, bool):
            raise TypeError('The current {} type is {}, expected int'.format(number, type(number).__name__))
        elif number < 0:
            return False
        return (number ** 0.5) % 1 == 0


class NumeralSystem:
    @staticmethod
    # TODO: Conversion from decimal to any other number system (from binary to thirty-six-bit)
    # TODO: Перевод из десятичной системы счисления в любую другую (от двоичной до тридцатишестиричной)
    def to_base(number: int, base: int, *, bits: int = -1) -> str:
        if not isinstance(number, int) or isinstance(number, bool):
            raise TypeError('The current {} type is {}, expected int'.format(number, type(number).__name__))
        elif not isinstance(base, int) or isinstance(base, bool):
            raise TypeError('The current {} type is {}, expected int'.format(base, type(base).__name__))
        elif not isinstance(bits, int) or isinstance(bits, bool):
            raise TypeError('The current {} type is {}, expected int'.format(bits, type(bits).__name__))
        elif not 2 <= base <= 36:
            raise ValueError('The base number system must be positive')
        elif bits != -1 and bits <= 0:
            raise ValueError('The number of bits must be positive')
        negative = False
        if number < 0:
            negative = True
            number = abs(number)
        alphabet = '0123456789abcdefghijklmnopqrstuvwxyz'
        result = alphabet[number % base]
        while number >= base:
            number //= base
            result += alphabet[number % base]
        return ('-' if negative else '') + result[::-1] if bits == -1 else \
            ('-' if negative else '') + (result + ('0' * (bits - len(result))))[:bits][::-1]


class PerformanceTracking:
    # TODO: Initialization of the class, the beginning of the countdown
    # TODO: Инициализация класса, начало отсчёта
    def __init__(self, *, expected: int = 60) -> None:
        if not isinstance(expected, int) or isinstance(expected, bool):
            raise TypeError('The current {} type is {}, expected int'.format(expected, type(expected).__name__))
        self.start = perf_counter()
        self.expected = expected

    # TODO: Object deletion, end of countdown
    # TODO: Удаление объекта, окончание отсчёта
    def __del__(self) -> None:
        result = perf_counter() - self.start
        print('\nThe program was completed in \033[1m{}{}\033[0m second!'.format('\033[32m'
                                                                                 if result <= self.expected
                                                                                 else '\033[31m',
                                                                                 result))

    @staticmethod
    # TODO: A decorated function for calculating the time of operation of the function
    # TODO: Декорируемая функция для вычисления времени работы функции
    def func_timer(*, expected: int = 10) -> Any:
        def decorator(func: Callable) -> Any:
            def wrapper(*args, **kwargs) -> Any:
                start = perf_counter()
                call = func(*args, **kwargs)
                result = perf_counter() - start
                print('Function «{}» was completed in \033[1m{}{}\033[0m second!\n'.format(func.__name__,
                                                                                           '\033[32m'
                                                                                           if result <= expected
                                                                                           else '\033[31m',
                                                                                           result))
                return call

            return wrapper

        return decorator



"""
:authors: notjik
:license: MIT License
:copyright: (c) 2023 notjik
"""
import os

from setuptools import setup, find_packages

version = '0.1.0'

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='scratches-algorithms',
    version=version,

    author='notjik',
    author_email='notjik@yandex.ru',

    description=(
        u'Python package with classical algorithms and simple utilities'
    ),
    long_description=long_description,
    long_description_content_type='text/markdown',

    url='https://github.com/notjik/scratches-algorithms',
    download_url='https://github.com/notjik/scratches-algorithms/dist/scratches-algorithms-0.1.0.tar.gz',

    license='MIT License',

    package_dir={'': 'src'},
    packages=find_packages('src', include=[
        'scratches_algorithms*'
    ]),

    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ]
)
